import napari
import os

from tifffile import imread

import pandas as pd



class TrainDataMaker(object):
       def __init__(self, source_dir):
              
              self.source_dir = source_dir
              self.acceptable_formats = [".tif", ".TIFF", ".TIF", ".png"]
              self._click_maker()

       def _click_maker(self):
           from qtpy.QtWidgets import QComboBox, QPushButton

           self.viewer = napari.Viewer()
           Imageids = []
           Boxname = 'ImageIDBox'
           X = os.listdir(self.source_dir)
           imageidbox = QComboBox()   
           imageidbox.addItem(Boxname)   
           tracksavebutton = QPushButton('Save Clicks')
           for imagename in X:
                  if any(imagename.endswith(f) for f in self.acceptable_formats):
                      Imageids.append(os.path.join(self.source_dir,imagename))
           for i in range(0, len(Imageids)):
                 imageidbox.addItem(str(Imageids[i])) 

           self.viewer.window.add_dock_widget(imageidbox, name="Image", area='bottom')    
           self.viewer.window.add_dock_widget(tracksavebutton, name="Save Clicks", area='bottom')
           imageidbox.currentIndexChanged.connect(
                        lambda trackid = imageidbox: EventViewer(
                                self.viewer,
                                imageidbox.currentText(),
                                os.path.basename(os.path.splitext(imageidbox.currentText())[0]), self.source_dir, False, True ))     

           tracksavebutton.clicked.connect(
                        lambda trackid= tracksavebutton:EventViewer(
                                self.viewer,
                                imageidbox.currentText(),
                                os.path.basename(os.path.splitext(imageidbox.currentText())[0]), self.source_dir, True, False ))                
                     
           napari.run()   


class EventViewer(object):

    def __init__(self, viewer, imagename, Name, csv_dir, save = False, newimage = True):
     
          self.save = save
          self.newimage = newimage
          self.viewer = viewer
          print('reading image')      
          self.imagename = imagename  
          self.image = imread(imagename)
          print('image read')
          self.Name = Name
          self.ndim = len(self.image.shape)  
          self.csv_dir = csv_dir
          assert self.ndim == 4, f'Input image should be 4 dimensional, try Training_data_maker for 2D + time images'
          
          self._click()

    def _click(self):

        
                ClassBackground = 'Normal'
                ClassMitosis = 'Mitosis'
                ClassApoptosis = 'Apoptosis'
                
                if self.save == True:
 
                        ClassBackgrounddata = self.viewer.layers[ClassBackground].data
                        bgdf = pd.DataFrame(ClassBackgrounddata, index = None, columns = ['T', 'Z', 'Y', 'X'])
                        bgdf.to_csv(self.csv_dir + '/' + 'ONEAT' + ClassBackground + self.Name +  '.csv', index = False, mode = 'w')

                        ClassMitosisdata = self.viewer.layers[ClassMitosis].data
                        divdf = pd.DataFrame(ClassMitosisdata, index = None, columns = ['T', 'Z', 'Y', 'X'])
                        divdf.to_csv(self.csv_dir + '/' +'ONEAT' + ClassMitosis + self.Name +  '.csv', index = False, mode = 'w')

                        ClassApoptosisdata = self.viewer.layers[ClassApoptosis].data
                        apopdf = pd.DataFrame(ClassApoptosisdata, index = None, columns = ['T', 'Z', 'Y', 'X'])
                        apopdf.to_csv(self.csv_dir + '/' +'ONEAT' + ClassApoptosis + self.Name +  '.csv', index = False, mode = 'w')

                       
        
                if self.newimage == True:
                     for layer in list(self.viewer.layers):

                            self.viewer.layers.remove(layer) 
                    
                if self.save == False:
                        self.viewer.add_image(self.image, name = self.Name)


                        # add the first points layer for ClassBackground Z and T point
                        self.viewer.add_points(name= ClassBackground, face_color='red', ndim = 4)

                        # add the second points layer for ClassMitosis Z and T point
                        self.viewer.add_points(name= ClassMitosis, face_color='blue', ndim = 4)

                        # add the second points layer for ClassApoptosis Z and T point
                        self.viewer.add_points(name= ClassApoptosis, face_color='green', ndim = 4)
                      
                        # programatically enter add mode for both Points layers to enable editing

                        self.viewer.layers[ClassBackground].mode = 'add'
                        self.viewer.layers[ClassMitosis].mode = 'add'
                        self.viewer.layers[ClassApoptosis].mode = 'add'


from tensorflow import keras as K
from tensorflow.keras import layers, models, regularizers, backend
from tensorflow.keras.layers import (
    Activation,
    BatchNormalization,
    Conv1D,
    Conv2D,
    Conv3D,
    ConvLSTM1D,
    ConvLSTM2D,
    concatenate
)
from tensorflow.keras.layers import Lambda
import os
reg_weight = 1.0e-4


class Concat(layers.Layer):
    def __init__(self, axis=-1, **kwargs):

        self.axis = axis
        self.kwargs = kwargs
        super().__init__()

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "axis": self.axis
            }
        )
        return config

    def call(self, x):

        y = Lambda(lambda x: layers.concatenate([x[0], x[1]], self.axis))(x)

        return y


def LRNet(
    input_shape,
    categories,
    box_vector,
    yolo_loss,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=32,
    input_model=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    depth of 29 == max pooling of 28 for image patch of 55
    depth of 56 == max pooling of 14 for image patch of 55
    """
    img_input = layers.Input(
        shape=(input_shape[0], None, None, input_shape[3])
    )
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    last_conv_factor = 2 ** (stage_number - 1)
    return_sequences = True
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_3d_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        return_sequences=return_sequences,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_3d_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                return_sequences=return_sequences,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = _resnet_3d_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                return_sequences=return_sequences,
                conv_first=False,
            )
            y = _resnet_3d_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                return_sequences=return_sequences,
                conv_first=False,
            )

            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = _resnet_3d_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    return_sequences=return_sequences,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = _resnet_3d_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )

    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid"
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid"
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
    

    return model

def _voll_dense_block(x, nb_layers,
    num_filters,
    kernel_size,
    activation):
    
    for _ in range(nb_layers):
        x = _voll_dense_conv(x, num_filters, kernel_size,activation)
    return x    
    
def _voll_transition_block(x,
                           reduction,
                           activation="relu"):
    
    x = BatchNormalization()(x)
    x = Activation(activation)(x)
    x = Conv3D(
        int(backend.int_shape(x)[-1] * reduction),
        1,
        use_bias=False
    )(x)
    
    x = layers.MaxPooling3D((2, 2, 2))(x)
    
    return x
    
def _voll_dense_conv(
    x,
    num_filters,
    kernel_size=3,
    activation="relu"
):
    
    
    y = BatchNormalization()(x)
    y = Activation(activation)(y)
    y = Conv3D(
        4 * num_filters,
        kernel_size= 1,
        use_bias = False,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(reg_weight),
    ) (y)
    y = BatchNormalization()(y)
    y = Activation(activation)(y)
    
    y = Conv3D(
        num_filters,
        kernel_size= kernel_size,
        use_bias = False,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(reg_weight),
    ) (y)
    
    block = Concat(-1)
    x = block([y, x])
    
    return x    
    
def _voll_conv(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    
    conv = Conv3D(
        num_filters,
        kernel_size=kernel_size,
        strides= strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def _voll_top(input_shape, stage_number):
    
        last_conv_factor = 2 ** (stage_number - 1) 
        print(input_shape, input_shape[0], input_shape[1], input_shape[2], input_shape[3])
        img_input = layers.Input(shape=(None, None, None, input_shape[3]))
        
        return last_conv_factor, img_input

def _voll_bottom(x, img_input, input_shape, categories, mid_kernel, last_conv_factor, last_activation, nboxes, box_vector, input_model, yolo_loss):
        
        x = (Conv3D(categories + nboxes * box_vector, kernel_size= mid_kernel, kernel_regularizer=regularizers.l2(reg_weight), padding = 'same'))(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)
        input_cat = Lambda(lambda x: x[:, :, :, :, 0:categories])(x)
        input_box = Lambda(lambda x: x[:, :, :, :, categories:])(x)
        output_cat = (
            Conv3D(
                categories,
                (
                    round(input_shape[0]/ last_conv_factor ),
                    round(input_shape[1] / last_conv_factor),
                    round(input_shape[2] / last_conv_factor),
                ),
                activation=last_activation,
                kernel_regularizer=regularizers.l2(reg_weight),
                padding="valid"
            )
        )(input_cat)
        output_box = (
            Conv3D(
                nboxes * (box_vector),
                (
                    round(input_shape[0] / last_conv_factor),
                    round(input_shape[1] / last_conv_factor),
                    round(input_shape[2] / last_conv_factor),
                ),
                activation="sigmoid",
                kernel_regularizer=regularizers.l2(reg_weight),
                padding="valid"
            )
        )(input_box)

        block = Concat(-1)
        outputs = block([output_cat, output_box])
        inputs = img_input
        
        # Create model.
        model = models.Model(inputs, outputs)

        if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

            model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
        return model
        
    
class DenseNet:
    def __init__(self, depth, startfilter, stage_number, start_kernel, mid_kernel,reduction, activation='relu'):
        self.stage_number = stage_number
        self.start_kernel = start_kernel
        self.mid_kernel = mid_kernel
        self.reduction = reduction
        self.depth = depth
        self.startfilter = startfilter
        self.activation = activation

    def __call__(self, x):
        self.nb_layers = []
        if type(self.depth) is dict:
            for (k,v) in self.depth.items():
                self.nb_layers.append(v) # get the list
                
        if len(self.nb_layers) != self.stage_number:
                raise ValueError('If `stage_number` is a list, its length must match '
                                 'the number of layers provided by `depth`.')
                
        x = _voll_conv(
            x,
            num_filters=self.startfilter,
            kernel_size=self.start_kernel,
            conv_first=True
        )        
        for stage in range(self.stage_number):
            
            x = _voll_dense_block(x, self.nb_layers[stage],
                self.startfilter//2,
                self.mid_kernel,
                self.activation)
            if stage < self.stage_number -1:
               x =_voll_transition_block(x,
                           self.reduction,
                           self.activation)
               
            x = BatchNormalization()(x)
            x = Activation(self.activation)(x)   
        
        return x
    
class ResNet:
    def __init__(self, depth, startfilter, stage_number, start_kernel, mid_kernel, activation='relu'):
        
        self.depth = depth
        self.startfilter = startfilter
        self.stage_number = stage_number
        self.start_kernel = start_kernel
        self.mid_kernel = mid_kernel
        self.activation = activation  
        
    def __call__(self, x):
        
        self.num_filters_in = self.startfilter
        if type(self.depth) is dict:
            for (k,v) in self.depth.items():
                self.nb_layers = v
        else:
                
              self.nb_layers = self.depth      
                
        num_res_blocks = int((self.nb_layers - 2) / 9)
        x = _voll_conv(
            x,
            num_filters=self.startfilter,
            kernel_size=self.start_kernel,
            conv_first=True
        )
              
        for stage in range(self.stage_number):
                for res_block in range(num_res_blocks):
                    activation = self.activation
                    batch_normalization = True
                    strides = 1
                    if stage == 0:
                        self.num_filters_out = self.num_filters_in * 4
                        if res_block == 0:  # first layer and first stage
                            activation = None
                            batch_normalization = False
                    else:
                        self.num_filters_out = self.num_filters_in * 2
                        if res_block == 0:  # not first layer and not first stage
                            strides = 2  # downsample

                    y = _voll_conv(
                        inputs=x,
                        num_filters=self.num_filters_in,
                        kernel_size=1,
                        strides=strides,
                        activation=activation,
                        batch_normalization=batch_normalization,
                        conv_first=False,
                    )
                    y = _voll_conv(
                        inputs=y,
                        num_filters=self.num_filters_in,
                        kernel_size=self.mid_kernel,
                        conv_first=False,
                    )
                    y = _voll_conv(
                        inputs=y,
                        num_filters=self.num_filters_out,
                        kernel_size=1,
                        conv_first=False,
                    )
                    if res_block == 0:
                        # linear projection residual shortcut connection to match
                        # changed dims
                        x = _voll_conv(
                            inputs=x,
                            num_filters=self.num_filters_out,
                            kernel_size=1,
                            strides=strides,
                            activation=None,
                            batch_normalization=False,
                        )

                    x = K.layers.add([x, y])
                self.num_filters_in = self.num_filters_out
                
        return x        
                
def DenseVollNet(
                input_shape,
                categories: dict,
                box_vector,
                yolo_loss,
                nboxes: int=1,
                start_kernel: int=7,
                mid_kernel: int=3,
                startfilter: int=64,
                stage_number: int = 3,
                input_model=None,
                last_activation: str="softmax",
                depth: dict={'depth_0': 6, 'depth_1': 12, 'depth_2':24, 'depth_3':16},
                reduction: float = 0.5
):
    
        last_conv_factor, img_input = _voll_top(input_shape = input_shape, stage_number = stage_number)
        densenet = DenseNet(depth, startfilter, stage_number, start_kernel, mid_kernel, reduction)
        x = densenet(img_input)
        model = _voll_bottom(x, img_input, input_shape, categories, mid_kernel, last_conv_factor, last_activation, nboxes, box_vector, input_model, yolo_loss)
        return model
        


def VollNet(
            input_shape,
            categories: dict,
            box_vector,
            yolo_loss,
            nboxes: int=1,
            start_kernel: int=7,
            mid_kernel: int=3,
            startfilter: int=64,
            stage_number: int = 3,
            input_model=None,
            last_activation: str="softmax",
            depth: dict={'depth_0': 29}
):
    
        last_conv_factor, img_input = _voll_top(input_shape = input_shape, stage_number = stage_number)
        resnet = ResNet(depth, startfilter, stage_number, start_kernel, mid_kernel)
        x = resnet(img_input)
        model = _voll_bottom(x, img_input, input_shape, categories, mid_kernel, last_conv_factor, last_activation, nboxes, box_vector, input_model, yolo_loss)
        return model


def resnet_lstm_v2(
    input_shape,
    categories,
    box_vector,
    yolo_loss,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_model=None,
    last_activation="softmax",
    
):
   

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(
        shape=(input_shape[0], input_shape[1], input_shape[2])
    )
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    return_sequences = True
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        return_sequences=return_sequences,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                return_sequences=return_sequences,
                conv_first=False,
            )
            y = _resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                return_sequences=return_sequences,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = _resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                return_sequences=return_sequences,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = _resnet_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    return_sequences=return_sequences,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = _resnet_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
     

    return model


def resnet_v2(
    input_shape,
    categories,
    box_vector,
    yolo_loss,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_model=None,
    last_activation="softmax",
):
   

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = _resnet_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = _resnet_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = _resnet_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
    

    return model


def resnet_1D_regression(
    input_shape,
    yolo_loss,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_model=None,
):
    """
    # Returns
        model (Model): Keras model instance
    """

    img_input = layers.Input(shape=input_shape)
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_layer_1D(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_layer_1D(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = _resnet_layer_1D(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = _resnet_layer_1D(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = _resnet_layer_1D(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (
        Conv2D(
            1,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)

    outputs = x

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
    

    return model


def resnet_v2_class(
    input_shape,
    categories,
    box_vector,
    yolo_loss,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_model=None,
    last_activation="softmax",
):
    

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = _resnet_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = _resnet_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = _resnet_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    outputs = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
    

    return model


def resnet_lstm_v2_class(
    input_shape,
    categories,
    box_vector,
    yolo_loss,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_model=None,
    last_activation="softmax",
):
   

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = _resnet_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            y = _resnet_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = _resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = _resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = _resnet_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling

    x = _resnet_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    outputs = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if any(fname.endswith('.pb') for fname in os.listdir(input_model)):

        model =  models.load_model(input_model,
                                custom_objects={'loss': yolo_loss, 'Concat': Concat})
    

    return model


def _resnet_3d_lstm_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    return_sequences=True,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):

    conv_lstm_3d = ConvLSTM2D(
        filters=num_filters,
        kernel_size=(kernel_size, kernel_size),
        activation=activation,
        strides=strides,
        data_format="channels_last",
        return_sequences=return_sequences,
        padding="same",
    )
    x = inputs
    if conv_first:
        x = conv_lstm_3d(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv_lstm_3d(x)
    return x


def _resnet_lstm_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    return_sequences=True,
    batch_normalization=True,
    conv_first=True,
):

    conv_lstm = ConvLSTM1D(
        filters=num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        return_sequences=return_sequences,
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv_lstm(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv_lstm(x)
    return x


def _resnet_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv2D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def _resnet_layer_1D(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    """1D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv1D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return 


try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
from csbdeep.utils.tf import keras_import

from oneat.NEATModels import NEATLRNet, NEATResNet, NEATTResNet, NEATVollNet
from oneat.pretrained import (
    clear_models_and_aliases,
    register_aliases,
    register_model,
)

get_file = keras_import("utils", "get_file")


clear_models_and_aliases(NEATVollNet, NEATLRNet, NEATTResNet, NEATResNet)

register_model(
    NEATLRNet,
    "Cellsplitdetectorbrightfield",
    "https://zenodo.org/record/6481021/files/Cellsplitdetectorbrightfield.h5",
    "0c6ba49c1ba0eb91819af40460fb66cb",
    "Cellsplitcordhelaflou",
    "https://zenodo.org/record/6481021/files/Cellsplitcordhelaflou.json",
    "aed21cb69d6fb8be32c47f78a39d32f5",
    "Cellsplitcategorieshelaflou",
    "https://zenodo.org/record/6481021/files/Cellsplitcategorieshelaflou.json",
    "7a67a83f08fb1add3c1b1a3e0eeec773",
    "Cellsplitdetectorbrightfield_Parameter",
    "https://zenodo.org/record/6481021/files/Cellsplitdetectorbrightfield_Parameter.json",
    "266e3e7fa7587d53d62ae0492482a1bc",
)


register_model(
    NEATVollNet,
    "Cellsplitdetectorhdpc",
    "https://zenodo.org/record/6483483/files/Cellsplitdetectorhdpc.h5",
    "701d7fc8494b66a73b024f3d3c3d3dad",
    "Cellsplitcordhelaflou",
    "https://zenodo.org/record/6481021/files/Cellsplitcordhelaflou.json",
    "aed21cb69d6fb8be32c47f78a39d32f5",
    "Cellsplitcategorieshelaflou",
    "https://zenodo.org/record/6481021/files/Cellsplitcategorieshelaflou.json",
    "7a67a83f08fb1add3c1b1a3e0eeec773",
    "Cellsplitdetectorhdpc_Parameter",
    "https://zenodo.org/record/6483483/files/Cellsplitdetectorhdpc_Parameter.json",
    "556f0ce063a4cacda6b9f27ae7aae69b",
)

register_model(
    NEATTResNet,
    "Cellsplitdetectorxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitdetectorxenopus.h5",
    "299a2edebf217da76732fd812cb5d6fe",
    "Cellsplitcordxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitcordxenopus.json",
    "aed21cb69d6fb8be32c47f78a39d32f5",
    "Cellsplitcategoriesxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitcategoriesxenopus.json",
    "7a67a83f08fb1add3c1b1a3e0eeec773",
    "Cellsplitdetectorxenopus_Parameter",
    "https://zenodo.org/record/6484966/files/Cellsplitdetectorxenopus_Parameter.json",
    "ee3dfb4ff1af80e44ceaebbdf4b1bff1",
)
register_model(
    NEATResNet,
    "Cellsplitdetectorxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitdetectorxenopus.h5",
    "299a2edebf217da76732fd812cb5d6fe",
    "Cellsplitcordxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitcordxenopus.json",
    "aed21cb69d6fb8be32c47f78a39d32f5",
    "Cellsplitcategoriesxenopus",
    "https://zenodo.org/record/6484966/files/Cellsplitcategoriesxenopus.json",
    "7a67a83f08fb1add3c1b1a3e0eeec773",
    "Cellsplitdetectorxenopus_Parameter",
    "https://zenodo.org/record/6484966/files/Cellsplitdetectorxenopus_Parameter.json",
    "ee3dfb4ff1af80e44ceaebbdf4b1bff1",
)

register_aliases(
    NEATLRNet, "Cellsplitdetectorbrightfield", "Cellsplitdetectorbrightfield"
)
register_aliases(NEATVollNet, "Cellsplitdetectorhdpc", "Cellsplitdetectorhdpc")
register_aliases(
    NEATTResNet, "Cellsplitdetectorxenopus", "Cellsplitdetectorxenopus"
)
register_aliases(
    NEATResNet, "Cellsplitdetectorxenopus", "Cellsplitdetectorxenopus"
)

del register_model, register_aliases, clear_models_and_aliases


def abspath(path):
    import os

    base_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(base_path, path)


def test_image_brightfield(target):
    import os

    from tifffile import imread, imwrite

    url = "https://zenodo.org/record/6480142/files/20210904_TL2%20-%20R05-C03-F0_ch_2.tif"
    hash = "67e13fa4df301dfe2c2a57f785aedada"
    fname = abspath(get_file(fname="brightfield", origin=url, file_hash=hash))
    image = imread(fname)
    Name = os.path.basename(os.path.splitext(fname)[0])
    imwrite(target + "/" + Name + ".tif", image.astype("float32"))
    return fname


def test_image_hdpc(target):
    import os

    from tifffile import imread, imwrite

    url = "https://zenodo.org/record/6483483/files/20210904_TL2%20-%20R05-C03-F0_ch_1.tif"
    hash = "ec6fde9a27a627866e88d44a6d98f41e"
    fname = abspath(get_file(fname="hdpc", origin=url, file_hash=hash))
    image = imread(fname)
    Name = os.path.basename(os.path.splitext(fname)[0])
    imwrite(target + "/" + Name + ".tif", image.astype("float32"))
    return fname


def test_image_xenopus(target):
    import os

    from tifffile import imread, imwrite

    url = "https://zenodo.org/record/6484966/files/C1-for_oneat_prediction.tif"
    hash = "ccf1836e66c350f90a3f458890d16220"
    fname = abspath(get_file(fname="xenopus", origin=url, file_hash=hash))
    image = imread(fname)
    Name = os.path.basename(os.path.splitext(fname)[0])
    imwrite(target + "/" + Name + ".tif", image.astype("float32"))
    return fname





class VizDet:
    
        
        
    def __init__(self):
        
        from oneat.NEATUtils.oneat_animation._qt import OneatVisualizeWidget
        import napari
        from qtpy.QtCore import Qt
        self.viewer = napari.Viewer()
        self.viz_widget = OneatVisualizeWidget(self.viewer)
        
        dock_widget = self.viewer.window.add_dock_widget(
            self.viz_widget, area="right"
        )
        self.viewer.window._qt_window.resizeDocks(
            [dock_widget], [200], Qt.Horizontal
        )

        napari.run()      
        


import numpy as np
import pandas as pd
from scipy import spatial


class ScoreModels:
    def __init__(
        self,
        predictions,
        groundtruth,
        thresholdscore=1 - 1.0e-4,
        thresholdspace=10,
        thresholdtime=2,
    ):

        # A list of all the prediction csv files, path object
        self.predictions = predictions
        # Approximate locations of the ground truth, Z co ordinate wil be ignored
        self.groundtruth = groundtruth
        self.thresholdscore = thresholdscore
        self.thresholdspace = thresholdspace
        self.thresholdtime = thresholdtime
        self.location_pred = []
        self.location_gt = []

        self.listtime_pred = []
        self.listy_pred = []
        self.listx_pred = []
        self.listscore_pred = []

        self.listtime_gt = []
        self.listy_gt = []
        self.listx_gt = []

    def model_scorer(self):

        Name = []
        TP = []
        FP = []
        FN = []
        GT = []
        Score_thresh = []
        Pred = []
        columns = [
            "Model Name",
            "True Positive",
            "False Positive",
            "False Negative",
            "Total Predictions",
            "GT predictions",
            "Score Threshold",
        ]

        dataset_gt = pd.read_csv(self.groundtruth, delimiter=",")
        T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
        Y_gt = dataset_gt[dataset_gt.keys()[2]][0:]
        X_gt = dataset_gt[dataset_gt.keys()[3]][0:]

        self.listtime_gt = T_gt.tolist()
        self.listy_gt = Y_gt.tolist()
        self.listx_gt = X_gt.tolist()
        for i in range(len(self.listtime_gt)):

            self.location_gt.append(
                [self.listtime_gt[i], self.listy_gt[i], self.listx_gt[i]]
            )

        for csv_pred in self.predictions:
            self.location_pred = []

            self.listtime_pred = []
            self.listy_pred = []
            self.listx_pred = []
            self.listscore_pred = []

            self.csv_pred = csv_pred
            name = self.csv_pred.stem
            dataset_pred = pd.read_csv(self.csv_pred, delimiter=",")
            T_pred = dataset_pred[dataset_pred.keys()[0]][0:]
            Y_pred = dataset_pred[dataset_pred.keys()[2]][0:]
            X_pred = dataset_pred[dataset_pred.keys()[3]][0:]
            Score_pred = dataset_pred[dataset_pred.keys()[4]][0:]

            self.listtime_pred = T_pred.tolist()
            self.listy_pred = Y_pred.tolist()
            self.listx_pred = X_pred.tolist()
            self.listscore_pred = Score_pred.tolist()

            for i in range(len(self.listtime_pred)):

                if float(self.listscore_pred[i]) >= float(self.thresholdscore):
                    self.location_pred.append(
                        [
                            int(self.listtime_pred[i]),
                            int(self.listy_pred[i]),
                            int(self.listx_pred[i]),
                        ]
                    )
            tp, fn, fp, pred, gt = self.TruePositives()

            Name.append(name)
            TP.append(tp)
            FN.append(fn)
            FP.append(fp)
            GT.append(gt)
            Pred.append(pred)
            Score_thresh.append(self.thresholdscore)
        data = list(zip(Name, TP, FP, FN, Pred, GT, Score_thresh))

        df = pd.DataFrame(data, columns=columns)
        df.to_csv(str(self.csv_pred.parent) + "_Model_Accuracy" + ".csv")
        return df

    def TruePositives(self):

        tp = 0
        fp = 0
        tree = spatial.cKDTree(self.location_gt)
        for i in range(len(self.location_pred)):

            return_index = self.location_pred[i]
            closestpoint = tree.query(return_index)
            spacedistance, timedistance = TimedDistance(
                return_index, self.location_gt[closestpoint[1]]
            )

            if (
                spacedistance < self.thresholdspace
                and timedistance < self.thresholdtime
            ):
                tp = tp + 1
            else:
                fp = fp + 1

        fn = self.FalseNegatives()
        return tp, fn, fp, len(self.location_pred), len(self.location_gt)

    def FalseNegatives(self):

        tree = spatial.cKDTree(self.location_pred)
        fn = 0
        for i in range(len(self.listtime_gt)):

            return_index = (
                int(self.listtime_gt[i]),
                int(self.listy_gt[i]),
                int(self.listx_gt[i]),
            )
            closestpoint = tree.query(return_index)
            spacedistance, timedistance = TimedDistance(
                return_index, self.location_pred[closestpoint[1]]
            )

            if (
                spacedistance > self.thresholdspace
                or timedistance > self.thresholdtime
            ):
                fn = fn + 1

        return fn


def TimedDistance(pointA, pointB):

    spacedistance = float(
        np.sqrt(
            (pointA[1] - pointB[1]) * (pointA[1] - pointB[1])
            + (pointA[2] - pointB[2]) * (pointA[2] - pointB[2])
        )
    )

    timedistance = float(np.abs(pointA[0] - pointB[0]))

    return spacedistance, timedistance



#!/usr/bin/env python3

import glob
import os
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from scipy.ndimage.filters import gaussian_filter
from tifffile import imread, imwrite


class Zmapgen:
    def __init__(
        self,
        imagedir,
        savedir,
        append_name="Mask",
        fileextension="*tif",
        show_after=1,
        radius=10,
    ):

        self.imagedir = imagedir
        self.savedir = savedir
        self.fileextension = fileextension
        self.show_after = show_after
        self.append_name = append_name
        self.radius = radius
        Path(self.savedir).mkdir(exist_ok=True)
        self.genmap()

    def genmap(self):

        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)
        count = 0
        for fname in X:
            Name = os.path.basename(os.path.splitext(fname)[0])
            if self.append_name in Name and "_Zmap" not in Name:
                image = imread(fname)
                count = count + 1
                Signal_first = image[:, :, :, 1]
                Signal_second = image[:, :, :, 2]
                Sum_signal_first = gaussian_filter(
                    np.sum(Signal_first, axis=0), self.radius
                )
                Sum_signal_first = normalizeZeroOne(Sum_signal_first)
                Sum_signal_second = gaussian_filter(
                    np.sum(Signal_second, axis=0), self.radius
                )

                Sum_signal_second = normalizeZeroOne(Sum_signal_second)

                Zmap = np.zeros(
                    [Sum_signal_first.shape[0], Sum_signal_first.shape[1], 3]
                )
                Zmap[:, :, 0] = Sum_signal_first
                Zmap[:, :, 1] = Sum_signal_second
                Zmap[:, :, 2] = (Sum_signal_first + Sum_signal_second) / 2
                if count % self.show_after == 0:
                    doubleplot(
                        Sum_signal_first,
                        Sum_signal_second,
                        Name + "First Channel Z map",
                        "Second Channel Z map",
                    )

                imwrite(self.savedir + Name + "_Zmap" + ".tif", Zmap)


def normalizeZeroOne(x):

    x = x.astype("float32")

    minVal = np.min(x)
    maxVal = np.max(x)

    x = (x - minVal) / (maxVal - minVal + 1.0e-20)

    return x


def doubleplot(imageA, imageB, titleA, titleB):
    fig, axes = plt.subplots(1, 2, figsize=(15, 6))
    ax = axes.ravel()
    ax[0].imshow(imageA, cmap=cm.Spectral)
    ax[0].set_title(titleA)

    ax[1].imshow(imageB, cmap=cm.Spectral)
    ax[1].set_title(titleB)

    plt.tight_layout()
    plt.show()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 30 16:05:34 2021
@author: vkapoor
"""

import tensorflow as tf
import numpy as np
from keras import backend as K

lambdaobject = 1
lambdanoobject = 1
lambdacoord = 1
lambdaclass = 1
lambdaangle = 1

def get_event_grid(grid_h, grid_w, grid_t, boxes):
    
    event_grid = np.array([ [[float(y),float(x),float(t)]]*boxes  for y in range(grid_h) for x in range(grid_w) for t in range(grid_t)])
    
    return event_grid


def extract_ground_event_truth(y_true, categories, grid_h, grid_w, grid_t, nboxes, box_vector):
    
    true_box_class = y_true[...,0:categories]
    
    true_nboxes = K.reshape(y_true[...,categories:], (-1, grid_h * grid_w * grid_t, nboxes, box_vector))

    true_box_xyt = true_nboxes[...,0:3]

    true_box_wh =  true_nboxes[...,3:5] 
    
    true_box_conf = true_nboxes[...,-1]
            
                
    return true_box_class, true_box_xyt, true_box_wh, true_box_conf
               
    
def extract_ground_event_pred(y_pred, categories, grid_h, grid_w, grid_t, event_grid, nboxes, box_vector):
    
    pred_box_class = y_pred[...,0:categories]
    
    pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w * grid_t, nboxes, box_vector))
    
    pred_box_xyt = pred_nboxes[...,0:3] + event_grid


    pred_box_wh = pred_nboxes[...,3:5]
        
    pred_box_conf = pred_nboxes[...,-1]
            
                
    return pred_box_class, pred_box_xyt, pred_box_wh, pred_box_conf  
            
    

def extract_ground_event_volume_truth(y_true, categories, grid_h, grid_w, grid_d, nboxes, box_vector):
    
    true_box_class = y_true[...,0:categories]
    
    true_nboxes = K.reshape(y_true[...,categories:], (-1, grid_h * grid_w * grid_d, nboxes, box_vector))
    
   

    true_box_xyz = true_nboxes[...,0:3]

    true_box_whd =  true_nboxes[...,4:7] 
    
    true_box_conf = true_nboxes[...,-1]
   
                
    return true_box_class, true_box_xyz, true_box_whd, true_box_conf 
               
    
def extract_ground_event_volume_pred(y_pred, categories, grid_h, grid_w, grid_d, event_grid, nboxes, box_vector):
    
    pred_box_class = y_pred[...,0:categories]
    
    pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w * grid_d, nboxes, box_vector))
    
    pred_box_xyz = pred_nboxes[...,0:3] + event_grid
    
    pred_box_whd = pred_nboxes[...,4:7]
        
    pred_box_conf = pred_nboxes[...,-1]
                
    return pred_box_class, pred_box_xyz, pred_box_whd, pred_box_conf 
            



def get_cell_grid(grid_h, grid_w, boxes):
    
    cell_grid = np.array([ [[float(x),float(y)]]*boxes   for y in range(grid_h) for x in range(grid_w)])
    
    return cell_grid

    
def extract_ground_cell_pred(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector):

        pred_box_class = y_pred[...,0:categories]
        
        pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        pred_box_xy = pred_nboxes[...,0:2] + (cell_grid)
    
        pred_box_wh = pred_nboxes[...,-3:-1]
        
        pred_box_conf = pred_nboxes[...,-1]
        
        return pred_box_class, pred_box_xy, pred_box_wh, pred_box_conf

def extract_ground_cell_truth(y_truth, categories, grid_h, grid_w, nboxes, box_vector):

        true_box_class = y_truth[...,0:categories]
        
        true_nboxes = K.reshape(y_truth[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        true_box_xy = true_nboxes[...,0:2]
    
        true_box_wh = true_nboxes[...,-3:-1]
        
        true_box_conf = true_nboxes[...,-1]
        
        return true_box_class, true_box_xy, true_box_wh, true_box_conf


def extract_ground_cell_pred_class(y_pred, categories):

        pred_box_class = y_pred[...,0:categories]
        

        return pred_box_class

def extract_ground_cell_pred_segfree(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector):

        pred_box_class = y_pred[...,0:categories]
        
        pred_nboxes = K.reshape(y_pred[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        pred_box_xy = pred_nboxes[...,0:2] + cell_grid
    
        
        return pred_box_class, pred_box_xy

def extract_ground_cell_truth_foc(y_truth, categories):

        true_box_class = y_truth[...,0:categories]
        
        return true_box_class

def extract_ground_cell_pred_foc(y_pred, categories):

        pred_box_class = y_pred[...,0:categories]
       
    
        
        return pred_box_class

def extract_ground_cell_truth_segfree(y_truth, categories, grid_h, grid_w, nboxes, box_vector):

        true_box_class = y_truth[...,0:categories]
        
        true_nboxes = K.reshape(y_truth[...,categories:], (-1, grid_h * grid_w, nboxes, box_vector))
    
        true_box_xy = true_nboxes[...,0:2]
    
        
        return true_box_class, true_box_xy

def extract_ground_cell_truth_class(y_truth, categories):

        true_box_class = y_truth[...,0:categories]
        
        return true_box_class
        
def compute_conf_loss(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf):
    
# compute the intersection of all boxes at once (the IOU)
        intersect_wh = K.maximum(K.zeros_like(pred_box_wh), (pred_box_wh + true_box_wh)/2 - K.square(pred_box_xy[...,0:1]- true_box_xy[...,0:1]) )
        intersect_area = intersect_wh[...,0] * intersect_wh[...,1] 
        true_area = true_box_wh[...,0] * true_box_wh[...,1] 
        pred_area = pred_box_wh[...,0] * pred_box_wh[...,1] 
        union_area = pred_area + true_area - intersect_area
        iou = tf.truediv(intersect_area , union_area)
  
        best_ious = K.max(iou, axis= -1)
        loss_conf = K.sum(K.square(true_box_conf*best_ious - pred_box_conf), axis=-1)

        loss_conf = loss_conf * lambdaobject

        return loss_conf 

def compute_conf_loss_volume(pred_box_whd, true_box_whd, pred_box_xyz,true_box_xyz,true_box_conf,pred_box_conf):
    
# compute the intersection of all boxes at once (the IOU)
        intersect_whd = K.maximum(K.zeros_like(pred_box_whd), (pred_box_whd + true_box_whd)/2 - K.square(pred_box_xyz[...,0:3]- true_box_xyz[...,0:3]) )
        intersect_area = intersect_whd[...,0] * intersect_whd[...,1] * intersect_whd[...,2] 
        true_area = true_box_whd[...,0] * true_box_whd[...,1] * true_box_whd[...,2]
        pred_area = pred_box_whd[...,0] * pred_box_whd[...,1] * pred_box_whd[...,2]
        union_area = pred_area + true_area - intersect_area
        iou = tf.truediv(intersect_area , union_area)
  
        best_ious = K.max(iou, axis= -1)
        loss_conf = K.sum(K.square(true_box_conf*best_ious - pred_box_conf), axis=-1)

        loss_conf = loss_conf * lambdaobject

        return loss_conf        

def compute_conf_loss_static(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf):
    
# compute the intersection of all boxes at once (the IOU)
        intersect_wh = K.maximum(K.zeros_like(pred_box_wh), (pred_box_wh + true_box_wh)/2 - K.square(pred_box_xy- true_box_xy) )
        intersect_area = intersect_wh[...,0] * intersect_wh[...,1] 
        true_area = true_box_wh[...,0] * true_box_wh[...,1] 
        pred_area = pred_box_wh[...,0] * pred_box_wh[...,1] 
        union_area = pred_area + true_area - intersect_area
        iou = intersect_area / union_area
  
        loss_conf = K.sum(K.square(true_box_conf*iou - pred_box_conf), axis=-1)

        loss_conf = loss_conf * lambdaobject

        return loss_conf        

def calc_loss_xywh(true_box_xy, pred_box_xy, true_box_wh, pred_box_wh):

    
    loss_xy      = K.sum(K.sum(K.square(true_box_xy - pred_box_xy), axis = -1), axis = -1)
    loss_wh      = K.sum(K.sum(K.square(K.sqrt(true_box_wh) - K.sqrt(pred_box_wh)), axis=-1), axis=-1)
    loss_xywh = (loss_xy + loss_wh)
    loss_xywh = lambdacoord * loss_xywh
    return loss_xywh

def calc_loss_xyzwhd(true_box_xyz, pred_box_xyz, true_box_whd, pred_box_whd):

    
    loss_xyz      = K.sum(K.sum(K.square(true_box_xyz - pred_box_xyz), axis = -1), axis = -1)
    loss_whd      = K.sum(K.sum(K.square(K.sqrt(true_box_whd) - K.sqrt(pred_box_whd)), axis=-1), axis=-1)
    loss_xyzwhd = (loss_xyz + loss_whd)
    loss_xyzwhd = lambdacoord * loss_xyzwhd
    return loss_xyzwhd    

def calc_loss_xy(true_box_xy, pred_box_xy):

    
    loss_xy      = K.sum(K.sum(K.square(true_box_xy - pred_box_xy), axis = -1), axis = -1)
    loss_xy = lambdacoord * loss_xy
    return loss_xy

def calc_loss_angle(true_box_angle, pred_box_angle):

    
    loss_angle      = K.sum(K.sum(K.square(true_box_angle - pred_box_angle), axis = -1), axis = -1)
    loss_angle = lambdaangle * loss_angle
    return loss_angle

def calc_loss_class(true_box_class, pred_box_class, entropy):

    
        loss_class   = K.mean(K.categorical_crossentropy(true_box_class, pred_box_class), axis=-1)

        loss_class   = loss_class * lambdaclass 

        return loss_class



def dynamic_yolo_loss(categories, grid_h, grid_w, grid_t, nboxes, box_vector, entropy):
    
    def loss(y_true, y_pred):    
        event_grid = get_event_grid(grid_h, grid_w, grid_t, nboxes)
        true_box_class, true_box_xyt, true_box_wh, true_box_conf = extract_ground_event_truth(y_true, categories, grid_h, grid_w,grid_t, nboxes, box_vector)
        pred_box_class, pred_box_xyt, pred_box_wh, pred_box_conf = extract_ground_event_pred(y_pred, categories, grid_h, grid_w,grid_t, event_grid, nboxes, box_vector)

        loss_xywht = calc_loss_xywh(true_box_xyt, pred_box_xyt, true_box_wh, pred_box_wh)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
        loss_conf = compute_conf_loss(pred_box_wh, true_box_wh, pred_box_xyt,true_box_xyt,true_box_conf,pred_box_conf) 
        combinedloss = (loss_xywht + loss_conf + loss_class) 


        return combinedloss 
        
    return loss 


def volume_yolo_loss(categories, grid_h, grid_w, grid_d, nboxes, box_vector, entropy, class_weights = None):
    def loss(y_true, y_pred):    
        event_grid = get_event_grid(grid_h, grid_w, grid_d, nboxes)
        true_box_class, true_box_xyz, true_box_whd, true_box_conf = extract_ground_event_volume_truth(y_true, categories, grid_h, grid_w,grid_d, nboxes, box_vector)
        pred_box_class, pred_box_xyz, pred_box_whd, pred_box_conf = extract_ground_event_volume_pred(y_pred, categories, grid_h, grid_w,grid_d, event_grid, nboxes, box_vector)

        loss_xywht = calc_loss_xyzwhd( true_box_xyz, pred_box_xyz, true_box_whd, pred_box_whd)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)

        loss_conf = compute_conf_loss_volume(pred_box_whd, true_box_whd, pred_box_xyz,true_box_xyz,true_box_conf,pred_box_conf)
                    # Adding it all up   
        if class_weights is None:             
            combinedloss = (loss_xywht + loss_conf + loss_class)
        else:
            # Apply class weights to the classification loss
            sample_weights = tf.reduce_sum(class_weights * tf.one_hot(tf.argmax(true_box_class, axis=-1), categories), axis=1)

            # Compute the weighted mean of the classification loss
            loss_class_weighted = tf.reduce_sum(loss_class * sample_weights) / tf.reduce_sum(sample_weights)

            # Adding it all up
            combinedloss = loss_xywht + loss_conf + loss_class_weighted

       

        return combinedloss 
        
    return loss





def static_yolo_loss(categories, grid_h, grid_w, nboxes, box_vector, entropy):
    
    def loss(y_true, y_pred):    

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        true_box_class, true_box_xy, true_box_wh, true_box_conf = extract_ground_cell_truth(y_true, categories, grid_h, grid_w, nboxes, box_vector, yolo_v0)
        pred_box_class, pred_box_xy, pred_box_wh, pred_box_conf = extract_ground_cell_pred(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector, yolo_v0)

        loss_xywh = calc_loss_xywh(true_box_conf, true_box_xy, pred_box_xy, true_box_wh, pred_box_wh)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
        loss_conf = compute_conf_loss_static(pred_box_wh, true_box_wh, pred_box_xy,true_box_xy,true_box_conf,pred_box_conf)
        
        combinedloss = (loss_xywh + loss_conf + loss_class)

        return combinedloss 
        
    return loss    

def static_yolo_loss_segfree(categories, grid_h, grid_w, nboxes, box_vector, entropy):
    
    def loss(y_true, y_pred):    

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        true_box_class, true_box_xy = extract_ground_cell_truth_segfree(y_true, categories, grid_h, grid_w, nboxes, box_vector)
        pred_box_class, pred_box_xy = extract_ground_cell_pred_segfree(y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector)
        loss_xy = calc_loss_xy(true_box_xy, pred_box_xy)

        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        
        combinedloss = (loss_xy + loss_class)

        return combinedloss 
        
    return loss  

def class_yolo_loss(categories, entropy):
    
    def loss(y_true, y_pred):    

      
        true_box_class = extract_ground_cell_truth_class(y_true, categories)
        pred_box_class = extract_ground_cell_pred_class(y_pred, categories)
        loss_class   = calc_loss_class(true_box_class, pred_box_class, entropy)
        combinedloss = loss_class

        return combinedloss 
        
    return loss    


import csv
import glob
import json
import os
from pathlib import Path

import numpy as np
import pandas as pd
from PIL import Image
from skimage import measure
from scipy import spatial
from sklearn.model_selection import train_test_split
from tifffile import imread, imwrite
from tqdm import tqdm

from .utils import normalizeFloatZeroOne


def SegFreeMovieLabelDataSet(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    normPatch=False,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])

        for csvfname in filesCsv:
            count = 0
            Csvname = os.path.basename(os.path.splitext(csvfname)[0])

            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                classfound = Csvname == csv_name_diff + event_name + name
                if classfound:
                    print(Csvname)
                    image = imread(fname).astype(dtype)
                    if not normPatch and normalizeimage:
                        image = normalizeFloatZeroOne(
                            image.astype(dtype), 1, 99.8, dtype=dtype
                        )
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]][1:]
                    y = dataset[dataset.keys()[1]][1:]
                    x = dataset[dataset.keys()[2]][1:]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():

                        SimpleMovieMaker(
                            t,
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                            normPatch,
                            tshift,
                        )
                        count = count + 1


def SegFreeMovieLabelDataSet4D(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    normPatch=False,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])

        for csvfname in filesCsv:
            count = 0
            Csvname = os.path.basename(os.path.splitext(csvfname)[0])

            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                classfound = Csvname == csv_name_diff + event_name + name
                if classfound:
                    print(Csvname)
                    image = imread(fname).astype(dtype)
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]]
                    z = dataset[dataset.keys()[1]]
                    y = dataset[dataset.keys()[2]]
                    x = dataset[dataset.keys()[3]]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():

                        SimpleMovieMaker4D(
                            normalizeimage,
                            t,
                            z[key],
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                            normPatch,
                            tshift,
                            dtype,
                        )
                        count = count + 1


def SimpleMovieMaker(
    time,
    y,
    x,
    image,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    normPatch,
    tshift,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time > 0:
        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            Event_data = []

            Label = np.zeros([total_categories + 6])

            newcenter = (y - shift[1], x - shift[0])
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                and time > size_tminus
                and time + size_tplus + 1 < image.shape[0]
            ):
                crop_xminus = x - int(imagesizex / 2)
                crop_xplus = x + int(imagesizex / 2)
                crop_yminus = y - int(imagesizey / 2)
                crop_yplus = y + int(imagesizey / 2)
                # Cut off the region for training movie creation
                region = (
                    slice(int(time - size_tminus), int(time + size_tplus + 1)),
                    slice(
                        int(crop_yminus) + shift[1], int(crop_yplus) + shift[1]
                    ),
                    slice(
                        int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]
                    ),
                )
                # Define the movie region volume that was cut
                crop_image = image[region]
                if normPatch:
                    crop_image = normalizeFloatZeroOne(
                        crop_image, 1, 99.8, dtype=dtype
                    )
                seglocationx = newcenter[1] - crop_xminus
                seglocationy = newcenter[0] - crop_yminus
                Label[total_categories] = seglocationx / sizex
                Label[total_categories + 1] = seglocationy / sizey
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                Label[total_categories + 3] = 1
                Label[total_categories + 4] = 1
                Label[total_categories + 5] = 1

                # Write the image as 32 bit tif file
                if (
                    crop_image.shape[0] == size_tplus + size_tminus + 1
                    and crop_image.shape[1] == imagesizey
                    and crop_image.shape[2] == imagesizex
                ):

                    imwrite(
                        (save_dir + "/" + newname + ".tif"),
                        crop_image.astype("float32"),
                    )
                    Event_data.append([Label[i] for i in range(0, len(Label))])
                    if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                        os.remove(save_dir + "/" + (newname) + ".csv")
                    writer = csv.writer(
                        open(save_dir + "/" + (newname) + ".csv", "a")
                    )
                    writer.writerows(Event_data)


def SimpleMovieMaker4D(
    normalizeimage,
    time,
    z,
    y,
    x,
    image,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    normPatch,
    tshift,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift

    image = image[:, z, :, :]
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )

    if time > 0:

        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            Event_data = []

            Label = np.zeros([total_categories + 6])
            Label[trainlabel] = 1

            newcenter = (y - shift[1], x - shift[0])
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                and time > size_tminus
                and time + size_tplus + 1 < image.shape[0]
            ):
                crop_xminus = x - int(imagesizex / 2)
                crop_xplus = x + int(imagesizex / 2)
                crop_yminus = y - int(imagesizey / 2)
                crop_yplus = y + int(imagesizey / 2)
                # Cut off the region for training movie creation
                region = (
                    slice(int(time - size_tminus), int(time + size_tplus + 1)),
                    slice(
                        int(crop_yminus) + shift[1], int(crop_yplus) + shift[1]
                    ),
                    slice(
                        int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]
                    ),
                )
                # Define the movie region volume that was cut
                crop_image = image[region]
                if normPatch:
                    crop_image = normalizeFloatZeroOne(
                        crop_image, 1, 99.8, dtype=dtype
                    )

                seglocationx = newcenter[1] - crop_xminus
                seglocationy = newcenter[0] - crop_yminus
                Label[total_categories] = seglocationx / sizex
                Label[total_categories + 1] = seglocationy / sizey
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                Label[total_categories + 3] = 1
                Label[total_categories + 4] = 1
                Label[total_categories + 5] = 1

                # Write the image as 32 bit tif file
                if (
                    crop_image.shape[0] == size_tplus + size_tminus + 1
                    and crop_image.shape[1] == imagesizey
                    and crop_image.shape[2] == imagesizex
                ):

                    imwrite(
                        (save_dir + "/" + newname + ".tif"),
                        crop_image.astype("float32"),
                    )
                    Event_data.append([Label[i] for i in range(0, len(Label))])
                    if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                        os.remove(save_dir + "/" + (newname) + ".csv")
                    writer = csv.writer(
                        open(save_dir + "/" + (newname) + ".csv", "a")
                    )
                    writer.writerows(Event_data)


def loadResizeImgs(im, size):

    w, h = im.size
    if w < h:
        im = im.crop((0, (h - w) / 2, w, (h + w) / 2))
    elif w > h:
        im = im.crop(((w - h + 1) / 2, 0, (w + h) / 2, h))

    return np.array(im.resize(size, Image.BILINEAR))


def Folder_to_oneat(
    dir, trainlabel, trainname, total_categories, size, save_dir
):

    Label = np.zeros([total_categories])

    count = 0

    files = sorted(glob.glob(dir + "/" + "*.png"))
    for i in tqdm(range(len(files))):
        file = files[i]
        try:
            Event_data = []

            img = Image.open(file)
            img = loadResizeImgs(img, size)
            Name = str(trainname) + os.path.basename(os.path.splitext(file)[0])
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            Label[trainlabel] = 1
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                image.astype("uint16"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)
            count = count + 1
        except Exception as e:
            print("[WW] ", str(e))
            continue


def Midog_to_oneat(
    midog_folder,
    annotation_file,
    event_type_name_label,
    all_ids,
    crop_size,
    save_dir,
):

    rows = []
    annotations = {}
    id_to_tumortype = {
        id: list(k for k in all_ids if id in all_ids[k])[0]
        for id in range(1, 406)
    }
    with open(annotation_file) as f:
        data = json.load(f)

        # categories = {cat["id"]: cat["name"] for cat in data["categories"]}
        total_categories = len(event_type_name_label.keys())
        for row in data["images"]:
            file_name = row["file_name"]
            image_id = row["id"]
            width = row["width"]
            height = row["height"]
            tumortype = id_to_tumortype[image_id]

            for annotation in [
                anno
                for anno in data["annotations"]
                if anno["image_id"] == image_id
            ]:
                box = annotation["bbox"]
                cat = annotation["category_id"]

                rows.append([image_id, width, height, box, cat, tumortype])
            annotations[file_name] = rows

    count = 0
    for tumortype, ids in zip(list(all_ids.keys()), list(all_ids.values())):

        for image_id in ids:

            file_path = midog_folder + "/" + f"{image_id:03d}.tiff"
            Name = os.path.basename(os.path.splitext(file_path)[0])

            img = imread(file_path)
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            image_annotation_array = annotations[Name + ".tiff"]
            image_id += 1

    for image_annotation in image_annotation_array:

        Label = np.zeros([total_categories + 5])
        Event_data = []
        (
            image_id,
            image_width,
            image_height,
            box,
            cat,
            tumortype,
        ) = image_annotation
        Name = str(image_id)
        x0, y0, x1, y1 = box
        height = y1 - y0
        width = x1 - x0
        x = (x0 + x1) // 2
        y = (y0 + y1) // 2
        # if cat == 1 then it is mitosis if cat == 2 it is hard negative
        if cat == 2:
            trainlabel = (
                event_type_name_label[tumortype] + total_categories // 2
            )
        if cat == 1:
            trainlabel = event_type_name_label[tumortype]
        ImagesizeX, ImagesizeY = crop_size
        crop_Xminus = x - int(ImagesizeX / 2)
        crop_Xplus = x + int(ImagesizeX / 2)
        crop_Yminus = y - int(ImagesizeY / 2)
        crop_Yplus = y + int(ImagesizeY / 2)
        region = (
            slice(int(crop_Yminus), int(crop_Yplus)),
            slice(int(crop_Xminus), int(crop_Xplus)),
        )

        crop_image = image[region]

        Label[trainlabel] = 1
        Label[total_categories] = 0.5
        Label[total_categories + 1] = 0.5
        Label[total_categories + 2] = width / ImagesizeX
        Label[total_categories + 3] = height / ImagesizeY
        Label[total_categories + 4] = 1

        count = count + 1
        if (
            crop_image.shape[0] == ImagesizeY
            and crop_image.shape[1] == ImagesizeX
        ):
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                crop_image.astype("float32"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)


def Midog_to_oneat_simple(
    midog_folder,
    annotation_file,
    event_type_name_label,
    all_ids,
    crop_size,
    save_dir,
):

    rows = []
    annotations = {}
    id_to_tumortype = {
        id: list(k for k in all_ids if id in all_ids[k])[0]
        for id in range(1, 406)
    }
    with open(annotation_file) as f:
        data = json.load(f)

        total_categories = len(event_type_name_label.keys())
        for row in data["images"]:
            file_name = row["file_name"]
            image_id = row["id"]
            width = row["width"]
            height = row["height"]
            tumortype = id_to_tumortype[image_id]

            for annotation in [
                anno
                for anno in data["annotations"]
                if anno["image_id"] == image_id
            ]:
                box = annotation["bbox"]
                cat = annotation["category_id"]

                rows.append([image_id, width, height, box, cat, tumortype])
            annotations[file_name] = rows

    count = 0
    for tumortype, ids in zip(list(all_ids.keys()), list(all_ids.values())):

        for image_id in ids:

            file_path = midog_folder + "/" + f"{image_id:03d}.tiff"
            Name = os.path.basename(os.path.splitext(file_path)[0])

            img = imread(file_path)
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            image_annotation_array = annotations[Name + ".tiff"]
            image_id += 1

    for image_annotation in image_annotation_array:

        Label = np.zeros([total_categories])
        Event_data = []
        (
            image_id,
            image_width,
            image_height,
            box,
            cat,
            tumortype,
        ) = image_annotation
        Name = str(image_id)
        x0, y0, x1, y1 = box
        height = y1 - y0
        width = x1 - x0
        x = (x0 + x1) // 2
        y = (y0 + y1) // 2
        # if cat == 1 then it is mitosis if cat == 2 it is hard negative
        if cat == 2:
            trainlabel = (
                event_type_name_label[tumortype] + total_categories // 2
            )
        if cat == 1:
            trainlabel = event_type_name_label[tumortype]
        ImagesizeX, ImagesizeY = crop_size
        crop_Xminus = x - int(ImagesizeX / 2)
        crop_Xplus = x + int(ImagesizeX / 2)
        crop_Yminus = y - int(ImagesizeY / 2)
        crop_Yplus = y + int(ImagesizeY / 2)
        region = (
            slice(int(crop_Yminus), int(crop_Yplus)),
            slice(int(crop_Xminus), int(crop_Xplus)),
        )

        crop_image = image[region]

        Label[trainlabel] = 1

        count = count + 1
        if (
            crop_image.shape[0] == ImagesizeY
            and crop_image.shape[1] == ImagesizeX
        ):
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                crop_image.astype("float32"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)


def MovieLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Seg_path = os.path.join(seg_image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesSeg = glob.glob(Seg_path)
    filesSeg.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:
        name = os.path.basename(os.path.splitext(fname)[0])
        for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                Csvname = csv_name_diff + event_name + name
                csvfname = os.path.join(csv_dir, Csvname + '.csv')
                if os.path.exists(csvfname):
                            print(Csvname)
                            count = 0
                            image = imread(os.path.join(image_dir,fname)).astype(dtype)
                            segimage = imread(os.path.join(seg_image_dir,fname)).astype("uint16")
                            dataset = pd.read_csv(csvfname)
                            time = dataset[dataset.keys()[0]]
                            y = dataset[dataset.keys()[1]]
                            x = dataset[dataset.keys()[2]]

                            # Categories + XYHW + Confidence
                            for (key, t) in time.items():

                                MovieMaker(
                                    t,
                                    y[key],
                                    x[key],
                                    image,
                                    segimage,
                                    crop_size,
                                    gridx,
                                    gridy,
                                    total_categories,
                                    trainlabel,
                                    name + event_name + str(count),
                                    save_dir,
                                    tshift,
                                    normalizeimage,
                                    dtype,
                                )
                                count = count + 1
                            image = None
                            segimage = None

def VolumeLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    gridz=1,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    files_raw =  os.listdir(image_dir)

    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:
        name = os.path.basename(os.path.splitext(fname)[0])
        for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                Csvname = csv_name_diff + event_name + name
                csvfname = os.path.join(csv_dir, Csvname + '.csv')
                if os.path.exists(csvfname):
                    print(Csvname)
                    count = 0
                    image = imread(os.path.join(image_dir,fname)).astype(dtype)
                    segimage = imread(os.path.join(seg_image_dir,fname)).astype("uint16")
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]]
                    z = dataset[dataset.keys()[1]]
                    y = dataset[dataset.keys()[2]]
                    x = dataset[dataset.keys()[3]]

                    # Categories + XYZHW + Confidence
                    for (key, t) in time.items():

                        VolumeMaker(
                            t,
                            z[key],
                            y[key],
                            x[key],
                            image,
                            segimage,
                            crop_size,
                            gridx,
                            gridy,
                            gridz,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                            tshift,
                            normalizeimage,
                            dtype,
                        )
                        count = count + 1
                    image = None
                    segimage = None



def VolumeMaker(
    time,
    z,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridx,
    gridy,
    gridz,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
    normalizeimage,
    dtype,
):

    sizex, sizey, sizez, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy
    imagesizez = sizez * gridz
    shiftNone = [0, 0, 0]

    AllShifts = [shiftNone]

    time = time - tshift
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )
    if time > size_tminus:

        # slice the images
    
        currentsegimage = segimage[int(time), :].astype("uint16")
        image_props = getHWD(
            x, y, z, currentsegimage, imagesizex, imagesizey, imagesizez
        )
        if image_props is not None:
            height, width, depth, center, seg_label = image_props
            smallimage = CreateVolume(
                image, size_tminus, size_tplus, int(time)
            )

            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                Event_data = []
                newcenter = center
                x = center[2]
                y = center[1]
                z = center[0]

                Label = np.zeros([total_categories + 8])
                Label[trainlabel] = 1

                # T co ordinate
                Label[total_categories + 3] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                if (
                    x > sizex / 2
                    and z > sizez / 2
                    and y > sizey / 2
                    and z + int(imagesizez / 2) < image.shape[1]
                    and y + int(imagesizey / 2) < image.shape[2]
                    and x + int(imagesizex / 2) < image.shape[3]
                    and time > size_tminus
                    and time + size_tplus + 1 < image.shape[0]
                ):
                    crop_xminus = x - int(imagesizex / 2)
                    crop_xplus = x + int(imagesizex / 2)
                    crop_yminus = y - int(imagesizey / 2)
                    crop_yplus = y + int(imagesizey / 2)
                    crop_zminus = z - int(imagesizez / 2)
                    crop_zplus = z + int(imagesizez / 2)
                    region = (
                        slice(0, smallimage.shape[0]),
                        slice(int(crop_zminus), int(crop_zplus)),
                        slice(int(crop_yminus), int(crop_yplus)),
                        slice(int(crop_xminus), int(crop_xplus)),
                    )

                    # Define the movie region volume that was cut
                    crop_image = smallimage[region]
                    seglocationx = newcenter[2] - crop_xminus
                    seglocationy = newcenter[1] - crop_yminus
                    seglocationz = newcenter[0] - crop_zminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    Label[total_categories + 2] = seglocationz / sizez
                    if height >= imagesizey:
                        height = 0.5 * imagesizey
                    if width >= imagesizex:
                        width = 0.5 * imagesizex
                    if depth >= imagesizez:
                        depth = 0.5 * imagesizez
                    # Height
                    Label[total_categories + 4] = height / imagesizey
                    # Width
                    Label[total_categories + 5] = width / imagesizex
                    # Depth
                    Label[total_categories + 6] = depth / imagesizez

                    Label[total_categories + 7] = 1
                    # Write the image as 32 bit tif file
                    if (
                        crop_image.shape[0] == size_tplus + size_tminus + 1
                        and crop_image.shape[1] == imagesizez
                        and crop_image.shape[2] == imagesizey
                        and crop_image.shape[3] == imagesizex
                    ):

                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)
    crop_image = None
    currentsegimage = None
    smallimage = None


def MovieMaker(
    time,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
    normalizeimage,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )
    if time > size_tminus:
        currentsegimage = segimage[int(time), :].astype("uint16")
        image_props = getHW(x, y, currentsegimage, imagesizex, imagesizey)
        if image_props is not None:
            height, width, center, seg_label = image_props
            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                Event_data = []
                newcenter = (center[0] - shift[1], center[1] - shift[0])
                x = center[1]
                y = center[0]

                Label = np.zeros([total_categories + 6])
                Label[trainlabel] = 1
                # T co ordinate
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                smallimage = CreateVolume(
                    image, size_tminus, size_tplus, int(time)
                )

                if (
                    x + shift[0] > sizex / 2
                    and y + shift[1] > sizey / 2
                    and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                    and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                    and time > size_tminus
                    and time + size_tplus + 1 < image.shape[0]
                ):
                    crop_xminus = x - int(imagesizex / 2)
                    crop_xplus = x + int(imagesizex / 2)
                    crop_yminus = y - int(imagesizey / 2)
                    crop_yplus = y + int(imagesizey / 2)
                    region = (
                        slice(0, smallimage.shape[0]),
                        slice(
                            int(crop_yminus) + shift[1],
                            int(crop_yplus) + shift[1],
                        ),
                        slice(
                            int(crop_xminus) + shift[0],
                            int(crop_xplus) + shift[0],
                        ),
                    )
                    # Define the movie region volume that was cut
                    crop_image = smallimage[region]

                    seglocationx = newcenter[1] - crop_xminus
                    seglocationy = newcenter[0] - crop_yminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    if height >= imagesizey:
                        height = 0.5 * imagesizey
                    if width >= imagesizex:
                        width = 0.5 * imagesizex
                    # Height
                    Label[total_categories + 3] = height / imagesizey
                    # Width
                    Label[total_categories + 4] = width / imagesizex
                    Label[total_categories + 5] = 1
                    # Write the image as 32 bit tif file
                    if (
                        crop_image.shape[0] == size_tplus + size_tminus + 1
                        and crop_image.shape[1] == imagesizey
                        and crop_image.shape[2] == imagesizex
                    ):

                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)
    image = None
    segimage = None
    crop_image = None
    currentsegimage = None
    smallimage = None


def Readname(fname):

    return os.path.basename(os.path.splitext(fname)[0])


def ImageLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    tshift=0,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Seg_path = os.path.join(seg_image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesSeg = glob.glob(Seg_path)
    filesSeg.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for csvfname in filesCsv:
        print(csvfname)
        count = 0
        Csvname = os.path.basename(os.path.splitext(csvfname)[0])

        for fname in files_raw:

            name = os.path.basename(os.path.splitext(fname)[0])
            for Segfname in filesSeg:

                Segname = os.path.basename(os.path.splitext(Segfname)[0])

                if name == Segname:

                    for i in range(0, len(static_name)):
                        event_name = static_name[i]
                        trainlabel = static_label[i]
                        if Csvname == csv_name_diff + name + event_name:
                            image = imread(fname).astype(dtype)
                            image = normalizeFloatZeroOne(
                                image.astype(dtype), 1, 99.8, dtype=dtype
                            )
                            segimage = imread(Segfname)
                            dataset = pd.read_csv(csvfname)
                            time = dataset[dataset.keys()[0]][1:]
                            y = dataset[dataset.keys()[1]][1:]
                            x = dataset[dataset.keys()[2]][1:]

                            # Categories + XYHW + Confidence
                            for (key, t) in time.items():
                                ImageMaker(
                                    t,
                                    y[key],
                                    x[key],
                                    image,
                                    segimage,
                                    crop_size,
                                    gridx,
                                    gridy,
                                    total_categories,
                                    trainlabel,
                                    name + event_name + str(count),
                                    save_dir,
                                    tshift,
                                )
                                count = count + 1


def SegFreeImageLabelDataSet(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for csvfname in filesCsv:
        print(csvfname)
        count = 0
        Csvname = os.path.basename(os.path.splitext(csvfname)[0])

        for fname in files_raw:

            name = os.path.basename(os.path.splitext(fname)[0])
            image = imread(fname).astype(dtype)

            image = normalizeFloatZeroOne(
                image.astype(dtype), 1, 99.8, dtype=dtype
            )
            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                if Csvname == csv_name_diff + name + event_name:
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]][1:]
                    y = dataset[dataset.keys()[1]][1:]
                    x = dataset[dataset.keys()[2]][1:]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():
                        SegFreeImageMaker(
                            t,
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                        )
                        count = count + 1


def CreateVolume(patch, size_tminus, size_tplus, timepoint):
    starttime = timepoint - int(size_tminus)
    endtime = timepoint + int(size_tplus) + 1
    smallimg = patch[starttime:endtime, :]

    return smallimg


def createNPZ(
    save_dir,
    axes,
    save_name="oneat",
    save_name_val="oneatVal",
    expand=True,
    static=False,
    flip_channel_axis=False,
    train_size=0.95,
):

    data = []
    label = []

    raw_path = os.path.join(save_dir, "*tif")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    NormalizeImages = [imread(fname) for fname in files_raw]

    names = [Readname(fname) for fname in files_raw]
    # Normalize everything before it goes inside the training
    for i in range(0, len(NormalizeImages)):

        n = NormalizeImages[i]

        blankX = n
        csvfname = save_dir + "/" + names[i] + ".csv"
        arr = []
        with open(csvfname) as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            arr = list(reader)[0]
            arr = np.array(arr)

        blankY = arr
        blankY = np.expand_dims(blankY, -1)
        if expand:

            blankX = np.expand_dims(blankX, -1)

        data.append(blankX)
        label.append(blankY)

    dataarr = np.asarray(data)
    labelarr = np.asarray(label)
    if flip_channel_axis:
        np.swapaxes(dataarr, 1, -1)
    if static:
        try:
            dataarr = dataarr[:, 0, :, :, :]
        except ValueError:
            pass
    print(dataarr.shape, labelarr.shape)
    traindata, validdata, trainlabel, validlabel = train_test_split(
        dataarr,
        labelarr,
        train_size=train_size,
        shuffle=False,
    )
    save_full_training_data(save_dir, save_name, traindata, trainlabel, axes)
    save_full_training_data(
        save_dir, save_name_val, validdata, validlabel, axes
    )


def _raise(e):
    raise e


def ImageMaker(
    time,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridX,
    gridY,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
):

    sizeX, sizeY = crop_size

    ImagesizeX = sizeX * gridX
    ImagesizeY = sizeY * gridY

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time < segimage.shape[0] - 1 and time > 0:
        currentsegimage = segimage[int(time), :].astype("uint16")

        image_props = getHW(x, y, currentsegimage, ImagesizeX, ImagesizeY)
        if image_props is not None:
            height, width, center, seg_label = image_props
            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                newcenter = (center[0] - shift[1], center[1] - shift[0])
                Event_data = []

                x = center[1]
                y = center[0]
                Label = np.zeros([total_categories + 5])
                Label[trainlabel] = 1
                if (
                    x + shift[0] > sizeX / 2
                    and y + shift[1] > sizeY / 2
                    and x + shift[0] + int(ImagesizeX / 2) < image.shape[2]
                    and y + shift[1] + int(ImagesizeY / 2) < image.shape[1]
                ):
                    crop_Xminus = x - int(ImagesizeX / 2)
                    crop_Xplus = x + int(ImagesizeX / 2)
                    crop_Yminus = y - int(ImagesizeY / 2)
                    crop_Yplus = y + int(ImagesizeY / 2)

                    for tex in range(int(time) - 2, int(time) + 2):
                        newname = newname + str(tex)
                        region = (
                            slice(int(tex - 1), int(tex)),
                            slice(
                                int(crop_Yminus) + shift[1],
                                int(crop_Yplus) + shift[1],
                            ),
                            slice(
                                int(crop_Xminus) + shift[0],
                                int(crop_Xplus) + shift[0],
                            ),
                        )

                        crop_image = image[region]

                        seglocationx = newcenter[1] - crop_Xminus
                        seglocationy = newcenter[0] - crop_Yminus

                        Label[total_categories] = seglocationx / sizeX
                        Label[total_categories + 1] = seglocationy / sizeY

                        if height >= ImagesizeY:
                            height = 0.5 * ImagesizeY
                        if width >= ImagesizeX:
                            width = 0.5 * ImagesizeX

                        Label[total_categories + 2] = height / ImagesizeY
                        Label[total_categories + 3] = width / ImagesizeX
                        Label[total_categories + 4] = 1

                        if (
                            crop_image.shape[1] == ImagesizeY
                            and crop_image.shape[2] == ImagesizeX
                        ):
                            imwrite(
                                (save_dir + "/" + newname + ".tif"),
                                crop_image.astype("float32"),
                            )
                            Event_data.append(
                                [Label[i] for i in range(0, len(Label))]
                            )
                            if os.path.exists(
                                save_dir + "/" + (newname) + ".csv"
                            ):
                                os.remove(save_dir + "/" + (newname) + ".csv")
                            writer = csv.writer(
                                open(save_dir + "/" + (newname) + ".csv", "a")
                            )
                            writer.writerows(Event_data)


def SegFreeImageMaker(
    time,
    y,
    x,
    image,
    crop_size,
    gridX,
    gridY,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
):

    sizex, sizey = crop_size

    ImagesizeX = sizex * gridX
    ImagesizeY = sizey * gridY

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time < image.shape[0] - 1 and time > 0:

        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            newcenter = (y - shift[1], x - shift[0])
            Event_data = []

            Label = np.zeros([total_categories + 5])
            Label[trainlabel] = 1
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(ImagesizeX / 2) < image.shape[2]
                and y + shift[1] + int(ImagesizeY / 2) < image.shape[1]
            ):
                crop_Xminus = x - int(ImagesizeX / 2)
                crop_Xplus = x + int(ImagesizeX / 2)
                crop_Yminus = y - int(ImagesizeY / 2)
                crop_Yplus = y + int(ImagesizeY / 2)

                for tex in range(int(time) - 2, int(time) + 2):
                    newname = newname + str(tex)
                    region = (
                        slice(int(tex - 1), int(tex)),
                        slice(
                            int(crop_Yminus) + shift[1],
                            int(crop_Yplus) + shift[1],
                        ),
                        slice(
                            int(crop_Xminus) + shift[0],
                            int(crop_Xplus) + shift[0],
                        ),
                    )

                    crop_image = image[region]

                    seglocationx = newcenter[1] - crop_Xminus
                    seglocationy = newcenter[0] - crop_Yminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    Label[total_categories + 2] = 1
                    Label[total_categories + 3] = 1
                    Label[total_categories + 4] = 1

                    if (
                        crop_image.shape[1] == ImagesizeY
                        and crop_image.shape[2] == ImagesizeX
                    ):
                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)


def getHW(defaultX, defaultY, currentsegimage, imagesizex, imagesizey):

    properties = measure.regionprops(currentsegimage)
    centroids = [prop.centroid for prop in properties]
    labels = [prop.label for prop in properties]
    tree = spatial.cKDTree(centroids)
    
    DLocation = (defaultY, defaultX)
    distance_cell_mask, nearest_location = tree.query(DLocation)
    if distance_cell_mask < 0.5 * imagesizex:
        y = int(centroids[nearest_location][0])         
        x = int(centroids[nearest_location][1])
        SegLabel = labels[nearest_location]
        DLocation = (y, x)
    else:
        if (
        int(TwoDLocation[0]) < currentsegimage.shape[0]
        and int(TwoDLocation[1]) < currentsegimage.shape[1]
        and all(i >= 0 for i in DLocation)
        ):
          SegLabel = currentsegimage[int(DLocation[0]), int(DLocation[1])]

        else:
            SegLabel = -1  

    TwoDLocation = (defaultY, defaultX)
    if (
        int(TwoDLocation[0]) < currentsegimage.shape[0]
        and int(TwoDLocation[1]) < currentsegimage.shape[1]
    ):
        SegLabel = currentsegimage[int(TwoDLocation[0]), int(TwoDLocation[1])]
        for prop in properties:

            if SegLabel > 0 and prop.label == SegLabel:
                minr, minc, maxr, maxc = prop.bbox
                center = (defaultY, defaultX)
                height = abs(maxc - minc)
                width = abs(maxr - minr)
                return height, width, center, SegLabel

            if SegLabel == 0:

                center = (defaultY, defaultX)
                height = 0.5 * imagesizex
                width = 0.5 * imagesizey
                return height, width, center, SegLabel


def getHWD(
    defaultX,
    defaultY,
    defaultZ,
    currentsegimage,
    imagesizex,
    imagesizey,
    imagesizez,
):

    properties = measure.regionprops(currentsegimage)
    centroids = [prop.centroid for prop in properties]
    labels = [prop.label for prop in properties]
    tree = spatial.cKDTree(centroids)
    
    DLocation = (defaultZ, defaultY, defaultX)
    distance_cell_mask, nearest_location = tree.query(DLocation)
    if distance_cell_mask < 0.5 * imagesizex:
        z = int(centroids[nearest_location][0])         
        y = int(centroids[nearest_location][1])
        x = int(centroids[nearest_location][2])
        SegLabel = labels[nearest_location]
        DLocation = (z, y, x)
    else:
        if (
        int(DLocation[0]) < currentsegimage.shape[0] 
        and int(DLocation[1]) < currentsegimage.shape[1]
        and int(DLocation[2]) < currentsegimage.shape[2]
        and all(i >= 0 for i in DLocation)
        ):
           SegLabel = currentsegimage[int(DLocation[0]), int(DLocation[1]), int(DLocation[2])]
        else:
            SegLabel = -1   
    if (
        int(DLocation[0]) < currentsegimage.shape[0]
        and int(DLocation[1]) < currentsegimage.shape[1]
        and int(DLocation[2]) < currentsegimage.shape[2]
    ):
       
        for prop in properties:
            if SegLabel > 0 and prop.label == SegLabel:
                minr, minc, mind, maxr, maxc, maxd = prop.bbox
                center = (defaultZ, defaultY, defaultX)
                height = abs(maxc - minc)
                width = abs(maxr - minr)
                depth = abs(maxd - mind)
                return height, width, depth, center, SegLabel

            if SegLabel == 0 :

                center = (defaultZ, defaultY, defaultX)
                height = 0.5 * imagesizex
                width = 0.5 * imagesizey
                depth = 0.5 * imagesizez
                return height, width, depth, center, SegLabel


def save_full_training_data(directory, filename, data, label, axes):
    """Save training data in ``.npz`` format."""

    len(axes) == data.ndim or _raise(ValueError())
    np.savez(directory + filename, data=data, label=label, axes=axes)


def InterchangeTXY(TXYCSV, save_dir):

    dataset = pd.read_csv(TXYCSV)
    time = dataset[dataset.keys()[0]][1:]
    x = dataset[dataset.keys()[1]][1:]
    y = dataset[dataset.keys()[2]][1:]

    Event_data = []

    Name = os.path.basename(os.path.splitext(TXYCSV)[0])

    for (key, t) in time.items():

        Event_data.append([t, y[key], x[key]])

    writer = csv.writer(open(save_dir + "/" + (Name) + ".csv", "a"))
    writer.writerows(Event_data)



import os

import napari
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget

from ..OneatVisualization import OneatVisualization
from .OneatFrameWidget import OneatFrameWidget


class OneatWidget(QWidget):
    """Widget for interatviely making oneat visualizations using the napari viewer.

    Parameters
    ----------
    viewer : napari.Viewer
        napari viewer.

    Attributes
    ----------
    key_frames : list of dict
        List of viewer state dictionaries.
    frame : int
        Currently shown key frame.
    """

    def __init__(
        self,
        viewer: "napari.viewer.Viewer",
        csvdir: str,
        savedir: str,
        savename: str,
        key_categories: dict,
        segimagedir: str = None,
        heatimagedir: str = None,
        heatname: str = "_Heat",
        start_project_mid: int = 0,
        end_project_mid: int = 1,
        event_count_plot: str = "Plot selected event count",
        cell_count_plot: str = "Plot cell count",
        event_norm_count_plot: str = "Plot selected normalized event count",
        parent=None,
    ):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._layout.addWidget(
            QLabel("Oneat Visualization Wizard", parent=self)
        )

        self.frameWidget = OneatFrameWidget(parent=self)
        self._layout.addWidget(self.frameWidget)

        self.start_prob = self.frameWidget.startprobSpinBox.value()
        self.nms_space = self.frameWidget.nmsspaceSpinBox.value()

        self.oneatvisualization = OneatVisualization(
            viewer,
            key_categories,
            csvdir,
            savedir,
            savename,
            self.frameWidget.ax,
            self.frameWidget.figure,
        )

        self.heatmapsteps = self.frameWidget.heatstepsSpinBox.value()
        self.event_threshold = float(self.frameWidget.label.text())

        self.frameWidget.plotidbox.addItem("Select a type of plot")
        self.frameWidget.plotidbox.addItem(event_count_plot)
        self.frameWidget.plotidbox.addItem(cell_count_plot)
        self.frameWidget.plotidbox.addItem(event_norm_count_plot)
        self.frameWidget.heatstepsSpinBox.valueChanged.connect(
            self.update_heat_steps
        )
        self.frameWidget.startprobSpinBox.valueChanged.connect(
            self.update_start_prob
        )
        self.frameWidget.nmsspaceSpinBox.valueChanged.connect(
            self.update_nms_space
        )

        self.frameWidget.scoreSlider.valueChanged.connect(self.updateLabel)

        self.frameWidget.imageidbox.currentIndexChanged.connect(
            lambda eventid=self.frameWidget.imageidbox: self._capture_image_callback(
                segimagedir,
                heatimagedir,
                heatname,
                start_project_mid,
                end_project_mid,
            )
        )

        self.frameWidget.eventidbox.currentIndexChanged.connect(
            lambda eventid=self.frameWidget.eventidbox: self._capture_csv_callback(
                segimagedir
            )
        )

        self.frameWidget.plotidbox.currentIndexChanged.connect(
            lambda eventid=self.frameWidget.imageidbox: self._capture_plot_callback(
                segimagedir,
                event_count_plot,
                cell_count_plot,
                event_norm_count_plot,
            )
        )

        self.frameWidget.recomputeButton.clicked.connect(
            lambda eventid=self.frameWidget.recomputeButton: self._start_callbacks(
                segimagedir,
                event_count_plot,
                cell_count_plot,
                event_norm_count_plot,
            )
        )

    def _start_callbacks(
        self,
        segimagedir,
        event_count_plot,
        cell_count_plot,
        event_norm_count_plot,
    ):

        self._capture_csv_callback(segimagedir)
        self._capture_plot_callback(
            segimagedir,
            event_count_plot,
            cell_count_plot,
            event_norm_count_plot,
        )

    def update_start_prob(self, event):
        """update state of 'heatmapsteps' at current key-frame to reflect GUI state"""
        self.start_prob = self.frameWidget.startprobSpinBox.value()

    def update_nms_space(self, event):
        """update state of 'heatmapsteps' at current key-frame to reflect GUI state"""
        self.nms_space = self.frameWidget.nmsspaceSpinBox.value()

    def update_heat_steps(self, event):
        """update state of 'heatmapsteps' at current key-frame to reflect GUI state"""
        self.heatmapsteps = self.frameWidget.heatstepsSpinBox.value()

    def updateLabel(self, value):

        real_value = float(
            self.start_prob + (1.0 - self.start_prob) / 5000 * float(value)
        )
        self.frameWidget.label.setText(str(real_value))
        self.event_threshold = float(real_value)

    def _capture_csv_callback(self, segimagedir):

        get_image_text = self.frameWidget.imageidbox.currentText()
        csv_event_name = self.frameWidget.eventidbox.currentText()
        imagename = os.path.basename(os.path.splitext(get_image_text)[0])
        self.oneatvisualization.show_csv(
            imagename,
            csv_event_name,
            segimagedir=segimagedir,
            event_threshold=self.event_threshold,
            heatmapsteps=self.heatmapsteps,
            nms_space=self.nms_space,
        )

    def _capture_image_callback(
        self,
        segimagedir,
        heatmapimagedir,
        heatname,
        start_project_mid,
        end_project_mid,
    ):

        get_image_text = self.frameWidget.imageidbox.currentText()
        imagename = os.path.basename(os.path.splitext(get_image_text)[0])
        self.oneatvisualization.show_image(
            get_image_text,
            imagename,
            segimagedir,
            heatmapimagedir,
            heatname,
            start_project_mid,
            end_project_mid,
        )

    def _capture_plot_callback(
        self,
        segimagedir,
        event_count_plot,
        cell_count_plot,
        event_norm_count_plot,
    ):

        plot_event_name = self.frameWidget.plotidbox.currentText()
        self.oneatvisualization.show_plot(
            plot_event_name,
            event_count_plot,
            event_norm_count_plot,
            cell_count_plot,
            segimagedir,
            self.event_threshold,
        )
